/**
   @file      age.h
   @author    Mitch Richling <richmit@member.ams.org>
   @Copyright Copyright 1998 by Mitch Richling.  All rights reserved.
   @brief     Include file for age report program.@EOL
   @Keywords  file age report multi-threaded
   @Std       C89

              This is the include file that creates the file age
              report program.  All the locking for the ageGbl*
              variables causes a serious bottleneck in this code;
              however, it is still relatively fast..
*/

/* *************************************************************************** */
/* Functionality options */

/* Call a function one each file */
#define FILEACTION 1
#define INITACTION 1
#define EXITACTION 1

/* Exit on not fatal, bad arguments. */
#define BADARGEXIT 0

/* Trivial Errors cause an exit. */
#define TRIVERREXIT 0

/* Print the list as we process it */
#define PRINTLIST 0

/* Collect statics */
#define DOSTATS 1

/* *************************************************************************** */
/* Tuning options */

/* Use memcpy instead of string functions. */
#define USEMEMCPY 0

/* Maximum number-1 of threads to run at one time. */
#define NUMTHREADS 55

/* Number of directories to read before updating global linked list. */
#define UPLOADTRIG 10

/* *************************************************************************** */
/* This is used for error messages. */
#define PROGNAME "age"

/* *************************************************************************** */
/* Global variables defined by this module. These are global to the main.c as well. */
#define NUMTBUCK 3
long int ageGblNumDirs[NUMTBUCK+2],   ageGblNumFiles[NUMTBUCK+2],  ageGblNumLinks[NUMTBUCK+2],         ageGblNumSpecial[NUMTBUCK+2];
long int ageGblTotalSize[NUMTBUCK+2], ageGblTotalBlks[NUMTBUCK+2], ageGblTotalRegFileSize[NUMTBUCK+2], ageGblTotalRegFileBlks[NUMTBUCK+2];
long int timeBuckets[NUMTBUCK] = {2592000, 5184000, 7776000};
char *buketNames[NUMTBUCK] = {"30 Days", "60 Days", "90 Days"};

/* Controls all the stats variables used by this code. */
pthread_mutex_t actionStats_mutex = PTHREAD_MUTEX_INITIALIZER;

/* *************************************************************************** */
/* This is meant to be overridden with useful stuff. */
int fileAction(struct stat s, char *name) {
  int buketIndex, i;

  pthread_mutex_lock(&actionStats_mutex);
  buketIndex = 0;
  for(i=0; (i<NUMTBUCK)&&(!buketIndex); i++)
    if(gblStartTime-timeBuckets[i] < s.st_mtime )
      buketIndex = i+1;
  if( (!buketIndex) && (gblStartTime-timeBuckets[NUMTBUCK-1] ))
    buketIndex = NUMTBUCK+1;
  ageGblTotalSize[buketIndex] += s.st_size;  
  ageGblTotalBlks[buketIndex] += s.st_blocks;
  if(S_ISDIR(s.st_mode)) {
    ageGblNumDirs[buketIndex]++;    
  } else if(S_ISREG(s.st_mode)) {
    ageGblNumFiles[buketIndex]++;
    ageGblTotalRegFileSize[buketIndex] += s.st_size;  
    ageGblTotalRegFileBlks[buketIndex]  += s.st_blocks;
  } else if(S_ISLNK(s.st_mode)) {
    ageGblNumLinks[buketIndex]++;
  } else { /* It must be special. :) */
    ageGblNumSpecial[buketIndex]++;
  }
  pthread_mutex_unlock(&actionStats_mutex);

  return (0);
}

/* *************************************************************************** */
/* This is meant to be overridden with useful stuff. */
int initAction(int argc, char *argv[], int *idxFirstArg) {
  int buketIndex;
  for(buketIndex=0; buketIndex<(NUMTBUCK+2); buketIndex++) {
    ageGblNumDirs[buketIndex]          = 0;
    ageGblNumFiles[buketIndex]         = 0;
    ageGblNumLinks[buketIndex]         = 0;
    ageGblNumSpecial[buketIndex]       = 0;
    ageGblTotalSize[buketIndex]        = 0;
    ageGblTotalBlks[buketIndex]        = 0;
    ageGblTotalRegFileSize[buketIndex] = 0;
    ageGblTotalRegFileBlks[buketIndex] = 0;
  }
  return (0);
}

/* *************************************************************************** */
/* This is meant to be overridden with useful stuff. */
int exitAction(void) {
  int buketIndex;
  char *name, *symb;
  for(buketIndex=0; buketIndex<(NUMTBUCK+2); buketIndex++) {
    printf("\n");
    if(buketIndex == 0) {
      name = "UNKNOWN";
      symb = " ";
    } else if(buketIndex==NUMTBUCK+1) {
      name = buketNames[NUMTBUCK-1];
      symb = ">";
    } else {
      name = buketNames[buketIndex-1];
      symb = "<";
    }
    printf("Bucket #%d %s%s\n", buketIndex, symb, name);
    printf("  Number of regular files...... %ld\n", ageGblNumFiles[buketIndex]);
    printf("  Number of links.............. %ld\n", ageGblNumLinks[buketIndex]);
    printf("  Number of directories........ %ld\n", ageGblNumDirs[buketIndex]);
    printf("  Number of special files...... %ld\n", ageGblNumSpecial[buketIndex]);
    printf("  Block Count.................. %ld\n", ageGblTotalBlks[buketIndex]);
    printf("  Total size................... %ld\n", ageGblTotalSize[buketIndex]);
    printf("  Total size (regular files)... %ld\n", ageGblTotalRegFileSize[buketIndex]);
    printf("  Block count (regular files).. %ld\n", ageGblTotalRegFileBlks[buketIndex]);
  }

  return 0;
}
