/**
   @file      csum.h
   @author    Mitch Richling <richmit@member.ams.org>
   @Copyright Copyright 1997 by Mitch Richling.  All rights reserved.
   @brief     This is the include file that creates the owner finder.@EOL
   @Keywords  none
   @Std       C89

              
*/


#include <ctype.h>
#include <errno.h>              /* error stf       POSIX */
#include <fcntl.h>              /* UNIX file ctrl  UNIX  */
#include <openssl/ssl.h>
#include <openssl/crypto.h>
#include <stdio.h>              /* I/O lib         ISOC  */
#include <stdlib.h>             /* Standard Lib    ISOC  */
#include <time.h>
#include <unistd.h>             /* UNIX std stf    POSIX */

/* *************************************************************************** */
/* Functionality options */

/* Call a function one each file */
#define FILEACTION 1
#define INITACTION 1
#define EXITACTION 0

/* Exit on not fatal, bad arguments. */
#define BADARGEXIT 0

/* Trivial Errors cause an exit. */
#define TRIVERREXIT 0

/* Print the list as we process it */
#define PRINTLIST 0

/* Collect statistics */
#define DOSTATS 0

/* *************************************************************************** */
/* Tuning options */

/* Use memcpy instead of string functions. */ 
#define USEMEMCPY 0

/* Maximum number-1 of threads to run at one time. */
#define NUMTHREADS 10

/* Number of directories to read before updating global linked list. */
#define UPLOADTRIG 10

/* *************************************************************************** */
/* This is used for error messages. */
#define PROGNAME "csum"

/* Number of bytes to read per read() call */
#define READ_SIZE 1024*128

/* Version of the output to deliver. */
#define OUTPUT_VERSION 2

/* Do we setup locks for OpenSSL? Near as I can tell, we don't need to
   setup locks for the OpenSSL library. I suspect the digest functions
   never use any of this stuff; however, the OpenSSL documentation
   isn't clear about WHAT parts of the library need locks. So I
   normally set this to 1 just to be safe... */
#define BE_OPENSSL_THREAD_SAFE 1

/* *************************************************************************** */
#if BE_OPENSSL_THREAD_SAFE
/* The array holding the locks for the OpenSSL lib.*/
pthread_mutex_t *theMutexesForSSL;
#endif

#if BE_OPENSSL_THREAD_SAFE
/** Callback function for OpenSSL library thread support */
void theLockCallBackForSSL(int mode, int n, const char *file, int line) {
  /*fprintf(stderr,"thread=%4d %s:%d\n", CRYPTO_thread_id(),file,line);*/
  if (mode & CRYPTO_LOCK) {
    pthread_mutex_lock(&(theMutexesForSSL[n]));
  } else {
    pthread_mutex_unlock(&(theMutexesForSSL[n]));
  }
}
#endif

/* *************************************************************************** */
/* This is meant to be overridden with useful stuff. */
int fileAction(struct stat s, char *name) {
  int FD;
  char fileBuf[READ_SIZE+10];
  char *nameCharPtr;
  int returnValue;
  unsigned long numLines    = 0;
  unsigned long numChars    = 0;
  unsigned long numBinChars = 0;
  EVP_MD_CTX md5ctx, sha1ctx;
  unsigned char md5val[EVP_MAX_MD_SIZE], sha1val[EVP_MAX_MD_SIZE];
  unsigned int md5len, i, sha1len;

  if(S_ISREG(s.st_mode)) {
    /* Open our file... */
    if((FD = open(name, O_RDONLY)) < 0) {
      perror("ERROR: File open");
      exit(10);
    } /* end if */

    /* Initialize the hash objects. */
    EVP_DigestInit(&md5ctx, EVP_md5());
    EVP_DigestInit(&sha1ctx, EVP_sha1());

    /* Read the file and feed data to the counters and hash functions. */
    while((returnValue = read(FD, fileBuf, READ_SIZE)) > 0) {
      for(i=0; i<returnValue; i++) {
        // Check for Binary chars
        if( !(isprint(fileBuf[i]) || isspace(fileBuf[i])) )
          numBinChars++;
        // Add up chars
        numChars++;
        // Check for NL
        if(fileBuf[i] == 10)
          numLines++;
      }
      EVP_DigestUpdate(&md5ctx, fileBuf, returnValue);
      EVP_DigestUpdate(&sha1ctx, fileBuf, returnValue);
    } /* end while */

    /* Make sure we exited well. */
    if(returnValue < 0) {
      perror("ERROR: File read");
      exit(11);
    } /* end if */

    /* We are done with the file, so we close it now. */
    if(close(FD) < 0) {
      perror("ERROR: File close");
      exit(13);
    } /* end if */

    /* Finish up the hash function computation. */
    EVP_DigestFinal(&md5ctx, md5val, &md5len);
    EVP_DigestFinal(&sha1ctx, sha1val, &sha1len);

    /* Print out the results */
    pthread_mutex_lock(&console_mutex);
    printf("%lu ", (unsigned long)time(NULL));
    if(OUTPUT_VERSION >= 2) {
      printf("%lu ", (unsigned long)s.st_atime);
      printf("%lu ", (unsigned long)s.st_ctime);
      printf("%lu ", (unsigned long)s.st_mtime);
    }
    printf("MD5:");
    for(i=0; i<md5len; i++)
      printf("%02x", md5val[i]);
    printf(" SHA1:");
    for(i=0; i<sha1len; i++)
      printf("%02x", sha1val[i]);
    printf(" %lu %lu %lu ", numBinChars, numLines, numChars);
    for(nameCharPtr=name; *nameCharPtr!='\0'; nameCharPtr++)
      if ( (*nameCharPtr==92) || (*nameCharPtr<33) || (*nameCharPtr>126) )
        printf("\\%03o", *nameCharPtr);
      else
        printf("%c", *nameCharPtr);
    printf("\n");
    pthread_mutex_unlock(&console_mutex);
  }

  return 0;  
}

/* *************************************************************************** */
/* This is meant to be overridden with useful stuff. */
int initAction(int argc, char *argv[], int *idxFirstArg) {
#if BE_OPENSSL_THREAD_SAFE
  int i;

#if defined(OPENSSL_THREADS)
  theMutexesForSSL=malloc(CRYPTO_num_locks() * sizeof(pthread_mutex_t));
  for (i=0; i<CRYPTO_num_locks(); i++)
    pthread_mutex_init(&(theMutexesForSSL[i]), NULL);
  CRYPTO_set_locking_callback(theLockCallBackForSSL);
#else
  printf("No SSL Thread Support!  Exiting!!\n");
  exit(1);
#endif

#endif
  return (0);
}

/* *************************************************************************** */
/* This is meant to be overridden with useful stuff. */
int exitAction(void) {
  return 0;
}
