/**
   @file      main_produce-consume.c
   @author    Mitch Richling <richmit@member.ams.org>
   @Copyright Copyright 1997,1998,1999 by Mitch Richling.  All rights reserved.
   @brief     @EOL
   @Keywords  none
   @Std       C89 POSIX UNIX98 (+S_ISLNK)

              I have used POSIX and ANSI C functions and macros where possible.  In some cases I have had
              to resort to UNIX98 as well.  The file access is mostly POSIX except for the handling of
              symbolic links.  The macro _POSIX_C_SOURCE is not defined as this would hide the macro for
              symbolic links on a compliant system.  This has been tested on SPARC Solaris 2.7, 2.8;
              x86_64 Solaris 10; MacOS X 10.1, 10.5, 10.6; RHEL 3.0, 4.6, 4.8; and SLES 10.2.

              A note on SAFEREADDIR: readir() uses a static structure to return results, but the POSIX
              standard requires that it use a DIFFERENT static structure per directory handle.  Thus
              readdir() COULD be POSIX compliant and thread safe in our application -- no two threads
              ever read the same directory.  Unfortunately the standard expressly allows readdir to be
              non-thread safe.  For example, it could use some other hidden global state across calls and
              still be in spec.  In practice most God fearing UNIX-like OSes don't do this, and thus have
              a perfectly safe readdir() for this application; however, if someone decides to use this
              code on a heathen UNIX-like OS we provide the option to use readdir_r via SAFEREADDIR and
              accept the extra code, malloc, and performance (lock) penalty.  Note taht many systems have
              a lock arround the innards of readdir_r that allow the process to only run one at a time,
              and that will dramatically slow things down when SAFEREADDIR is 1.

             Functionality:

              This program traverses a file system and performs actions on that file system.  It is
              depth-first in it's traversal mechanism.  It is multi-threaded as well.  On very large
              directories it can be memory hungry--less than 100Mb.  It has the ability to print out
              various bits of information.  It is primarily customizable through the use of #define
              directives.
*/

/* DEBUG is used to turn on lots of debugging output to stdout. */
#ifndef DEBUG
#define DEBUG 0
#endif

/* *************************************************************************** */
#include <sys/types.h>          /* UNIX types.       */
#include <stdio.h>              /* ANSI C I/O lib    */
#include <string.h>             /* ANSI C strings    */
#include <dirent.h>             /* UNIX directories  */
#include <errno.h>              /* error functions   */
#include <utime.h>              /* utime             */
#include <sys/stat.h>           /* lstat             */
#include <time.h>               /* time              */
#include <pthread.h>            /* POSIX threads     */
#include <sched.h>              /* POSIX threads     */
#include <stdlib.h>             /* malloc            */
#include <unistd.h>             /* Std UNIX stuff    */
#include <stddef.h>

/* *************************************************************************** */
/* The structure that holds file information in the linked list.  This
   is intended to be customized for different applications. */
struct myStatStruct {
  char *d_name;
  struct myStatStruct *nextNode;
};

/* Report status every STATSECS seconds -- 0 to turn it off. */
#define STATSECS 10

/* Hardwire support for skipping .snapshot directories (NetApp'ism) */
#ifndef SKIPSNAPSHOT
#define SKIPSNAPSHOT 0
#endif

/* Use absolutely thread-safe readdir_r function, or use the much
   faster, mostly thread-safe readdir system call */
#ifndef SAFEREADDIR
#define SAFEREADDIR 0
#endif

/* *************************************************************************** */
/* Various mutexes for the threads. */

/* Controls all the stats variables */
pthread_mutex_t stats_mutex = PTHREAD_MUTEX_INITIALIZER;

/* Controls access to gblWorkingDir and gblLastDir(the global linked list) */
pthread_mutex_t dirList_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t dirList_cond = PTHREAD_COND_INITIALIZER;

/* Controls access to STDOUT.  Keeps console prints from clobbering each other */
pthread_mutex_t console_mutex = PTHREAD_MUTEX_INITIALIZER;

/* *************************************************************************** */
/* Global variables required for threaded operation. They are defined
   here so that the functions defined in the headers can use them. */

/* These form the basis of the global linked list. */
struct myStatStruct *gblLastDir = NULL, *gblWorkingDir = NULL;

/* All the global stats variables.  These are updated only when each
   directory reading thread ends.  These may not be used if we don't
   keep stats, but they are here anyhow. They should be in the header,
   or the header Should be split up into two parts. */
long int gblNumDirs   = 1, gblNumFiles  = 0, gblNumLinks         = 0, gblNumSpecial       = 0;
long int gblTotalSize = 0, gblTotalBlks = 0, gblTotalRegFileSize = 0, gblTotalRegFileBlks = 0;
time_t   gblStartTime = 0, gblEndTime   = 0;

/* *************************************************************************** */
/* Prototypes. */

int main(int argc, char *argv[]);

/* Designed to be used as a thread. */
void readDir(struct myStatStruct *workingDir);

/* This is a worker thread function that is intended to run readDir as a function. */
void workerThread();

/* *************************************************************************** */
/* These include files house the bits that are different for each binary. */
#if defined (dirgo_uidgidch) || defined (dirgo_uidch) || defined (dirgo_chown) || defined (dirgo_ltsstage)
#include "uidgidch.h"
#elif defined (dirgo_dux) || (dirgo_du)
#include "du.h"
#elif defined (dirgo_noown) || (dirgo_own)
#include "own.h"
#elif defined (dirgo_age)
#include "age.h"
#elif defined (dirgo_csum)
#include "csum.h"
#elif defined (dirgo_ownRep)
#include "ownRep.h"
#elif defined (dirgo_all)
#include "all.h"
#endif

/* Customizable function that is used to perform file actions.  These
   are here to keep the prototypes consistent. */
#if INITACTION
int initAction(int argc, char *argv[], int *idxFirstArg);
#endif
#if FILEACTION
int fileAction(struct stat s, char *name);
#endif
#if EXITACTION
int exitAction(void);
#endif

int numThreadsWaiting;

/* *************************************************************************** */
/* This is where it starts. :) */
int main(int argc, char *argv[]) {
  struct stat s;
  struct myStatStruct *tmpDir;
  pthread_t thread;
  int mustWait = 1, i;
  int idxFirstArg = 1;
  time_t statusTriger;
  struct timespec nanotime;

#if DOSTATS
  gblStartTime = time(NULL);
#endif

#if INITACTION
  initAction(argc, argv, &idxFirstArg);
#endif

  /* check and process command line arguments. */
  if((argc - idxFirstArg) < 1) {
    printf("ERROR(10): %s: usage : %s\n", PROGNAME, "Too Few Arguments");
    exit(1);
  } /* end if */

  /* Walk through the rest of the arguments and populate the list of
     directories and take file actions.  We don't have to use mutex
     locks here as we have spawned no threads yet. */
  gblWorkingDir = NULL;
  for(i = idxFirstArg; i < argc; i++) {
#if DEBUG >= 6
  printf("Arg Process(%d): %s\n", i, argv[i]);
#endif
    /* First we stat the directory. */
    if(lstat(argv[i], &s)) {
      printf("ERROR(1): %s: lstat: %s: %s\n", PROGNAME, argv[i], "Undetermined");
      if(TRIVERREXIT)
        exit(1);
    } else {
      if(FILEACTION && !(S_ISDIR(s.st_mode))) {
#if FILEACTION
        fileAction(s, argv[i]);
#endif
      } else {
        /* Get RAM for the data structures. */
        if((tmpDir = (struct myStatStruct *)malloc(sizeof(struct myStatStruct))) == NULL) {
          printf("ERROR(2): %s: malloc: %s\n", PROGNAME, "Undetermined");
          exit(1);
        } /* end if */
        if((tmpDir->d_name = (char *)malloc(sizeof(argv[i][0]) * (strlen(argv[i]) + 1))) == NULL) {
          printf("ERROR(3): %s: malloc: %s\n", PROGNAME, "Undetermined");
          exit(1);
        } /* end if */
        /* Populate the data structure and the linked list. */
        strcpy(tmpDir->d_name, argv[i]);
        tmpDir->nextNode = NULL;
        if(gblWorkingDir == NULL)
          gblWorkingDir = tmpDir;
        else
          gblLastDir->nextNode = tmpDir;
        gblLastDir = tmpDir;
#if FILEACTION
        fileAction(s, tmpDir->d_name);
#endif
      } /* end if/else */
    } /* end if/else */
  } /* end for */
  
  /* Lock dirList to keep threads from running.. */
  pthread_mutex_lock(&dirList_mutex);
#if DEBUG >= 6
  printf("About to kick off threads...\n");
#endif
  numThreadsWaiting = 0;
  /* Here we fire off n-threads */
  for(i = 0; i < NUMTHREADS; i++) {
    if(pthread_create(&thread, NULL, (void *(*)(void *))workerThread, (void *)&i) != 0) {
      pthread_mutex_lock(&console_mutex);
      printf("ERROR(9): %s: pthread_create: %s\n", PROGNAME, "Couldn't create thread.");
      pthread_mutex_unlock(&console_mutex);
      exit(1);
    } /* end if */
  } /* end for */
  pthread_mutex_unlock(&dirList_mutex);

  mustWait = 1;
  statusTriger = time(NULL);
  while(mustWait) {
    pthread_mutex_lock(&dirList_mutex);
    if((gblWorkingDir == NULL) && (numThreadsWaiting == NUMTHREADS)) {
      mustWait = 0;
    } /* end if */
    pthread_mutex_unlock(&dirList_mutex);

    pthread_cond_broadcast(&dirList_cond);

    nanotime.tv_sec = 0;
    nanotime.tv_nsec = 1000;
    nanosleep(&nanotime, NULL);

#if STATSECS != 0
    if( (time(NULL)-statusTriger) > STATSECS) {
      statusTriger = time(NULL);
      /* We don't lock at this point because it's OK if this is not printed out nicely... */
#if DOSTATS
      //pthread_mutex_lock(&stats_mutex);
      //pthread_mutex_lock(&console_mutex);
      printf("Number of file-system objects found: %ld\n", 
             (long)(gblNumDirs + gblNumDirs + gblNumFiles + gblNumLinks + gblNumSpecial));
      //pthread_mutex_unlock(&console_mutex);
      //pthread_mutex_unlock(&stats_mutex);
#else
      //pthread_mutex_lock(&console_mutex);
      //printf("Working\n");
      //pthread_mutex_unlock(&console_mutex);
#endif
    } /* end if */
#endif
  } /* end while */
  pthread_mutex_unlock(&dirList_mutex);

#if DOSTATS
  gblEndTime = time(NULL);
#endif

  /* Time to exit. */
#if EXITACTION
  exitAction();
#endif
  return (0);
} /* end func main */

/* *************************************************************************** */
void workerThread() {
  struct myStatStruct *workingDir;

#if DEBUG >= 10
  pthread_mutex_lock(&console_mutex);
  printf("Thread Startup.\n");
  pthread_mutex_unlock(&console_mutex);
#endif

  /* We are just waiting till we can lock to keep everyone in sync. */
  pthread_mutex_lock(&dirList_mutex);
  pthread_mutex_unlock(&dirList_mutex);

  while(1) {
    pthread_mutex_lock(&dirList_mutex);
    if(gblWorkingDir == NULL) {
      numThreadsWaiting++;

      while(gblWorkingDir == NULL) {
        pthread_cond_wait(&dirList_cond, &dirList_mutex);
      }                         /* If we get here, we have dirList_mutex locked. */
      numThreadsWaiting--;
      pthread_mutex_unlock(&dirList_mutex);
    } else {
      workingDir = gblWorkingDir;
      gblWorkingDir = gblWorkingDir->nextNode;
      pthread_mutex_unlock(&dirList_mutex);
      readDir(workingDir);
    } /* end if/else */
  } /* end while */
} /* end func workerThread */

/* *************************************************************************** */
void readDir(struct myStatStruct *workingDir) {
  DIR *dp;
  struct stat s;
  char *name;
  long len_d_name, len_workingDirName;
  struct myStatStruct *firstDir = NULL, *lastDir = NULL, *tmpDir;
  struct dirent *dep;
#if SAFEREADDIR
  struct dirent *depp;
  int readdirRes;
  size_t depSize;
#endif
  int haveMore, name_max;

#if UPLOADTRIG > 0
  int uploadTrig = 0;
#endif
#if DOSTATS
  long int numDirs   = 1, numFiles  = 0, numLinks         = 0, numSpecial       = 0;
  long int totalSize = 0, totalBlks = 0, totalRegFileSize = 0, totalRegFileBlks = 0;
#endif

#if DEBUG >= 6
  pthread_mutex_lock(&console_mutex);
  printf("     Dir Rec: %s\n", workingDir->d_name);
  pthread_mutex_unlock(&console_mutex);
#endif

  dp = opendir(workingDir->d_name);
  if(dp == NULL) {
    printf("ERROR(4): %s: opendir: %s: %s\n", PROGNAME, workingDir->d_name, "Undetermined");
    if(TRIVERREXIT)
      exit(1);
  } else {
    name_max=pathconf(workingDir->d_name, _PC_NAME_MAX);
    if(name_max <= 0) {
      printf("ERROR(7): %s: pathconf: %s: Undetermined\n", PROGNAME, workingDir->d_name);
      exit(1);
    } /* end if */

#if SAFEREADDIR
    /* Some platforms have large dirent sizes...  Broken if d_name not last thing... */
    depSize = offsetof(struct dirent, d_name) + name_max + 2;
    if (depSize < sizeof(struct dirent))
      depSize = sizeof(struct dirent);
    if((dep = (struct dirent *)malloc(depSize)) == NULL) {
      printf("ERROR(8): %s: malloc: %s\n", PROGNAME, "dirent+");
      exit(1);
    } /* end if */
#endif

    len_workingDirName = strlen(workingDir->d_name);
    haveMore = 1;
    while(haveMore) {
#if SAFEREADDIR
      readdirRes = readdir_r(dp, dep, &depp);
#else
      dep = readdir(dp);
#endif
      if(
#if SAFEREADDIR
        /* Note that we should do something diffrent for errors for readdir_r... */
        ((depp == NULL) || (readdirRes != 0))
#else
        /* Note that we should check errno for readdir as it might not be EOF but an error... */
        (dep == NULL)
#endif
        ) {
        haveMore = 0;
      } else {
        if((strcmp(dep->d_name, "..") != 0) && (strcmp(dep->d_name, ".") != 0) 
#if SKIPSNAPSHOT
           && (strcmp(dep->d_name, ".snapshot") != 0)
#endif
          ) {
          len_d_name = strlen(dep->d_name);
          if((name = (char *)malloc(len_d_name + len_workingDirName + 2)) == NULL) {
            printf("ERROR(5): %s: malloc: %s\n", PROGNAME, "Undetermined");
            exit(1);
          } /* end if */

          memcpy(name, workingDir->d_name, len_workingDirName);
          name[len_workingDirName] = '/';
          memcpy(name + (len_workingDirName + 1), dep->d_name, len_d_name + 1);

          if(lstat(name, &s)) {
            printf("ERROR(6): %s: lstat: %s: %s\n", PROGNAME, name, "Undetermined");

#if DEBUG >= 10
            pthread_mutex_lock(&console_mutex);
            pthread_mutex_lock(&dirList_mutex);
            printf("=Global List=========================================\n");
            tmpDir = gblWorkingDir;
            while(tmpDir != NULL) {
              printf(":%s:\n", tmpDir->d_name);
              tmpDir = tmpDir->nextNode;
            } /* end while */
            printf("====================================================\n");
            pthread_mutex_unlock(&dirList_mutex);
            pthread_mutex_unlock(&console_mutex);
#endif

            if(TRIVERREXIT)
              exit(1);
          } else {                /* We have stated it and want to do something */
#if ( (PRINTLIST > 0) && (DEBUG < 5) )
            pthread_mutex_lock(&console_mutex);
            printf("%s\n", name);
            pthread_mutex_unlock(&console_mutex);
#endif
            if(S_ISDIR(s.st_mode)) {      /* Process directories. */
              if((tmpDir = (struct myStatStruct *)malloc(sizeof(struct myStatStruct))) == NULL) {
                printf("ERROR(7): %s: malloc: %s\n", PROGNAME, "Undetermined");
                exit(1);
              } /* end if */
              if((tmpDir->d_name = (char *)malloc(len_d_name + len_workingDirName + 2)) == NULL) {
                printf("ERROR(8): %s: malloc: %s\n", PROGNAME, "Undetermined");
                exit(1);
              } /* end if */
              strcpy(tmpDir->d_name, name);
              tmpDir->nextNode = NULL;

              if(lastDir == NULL) {
                firstDir = tmpDir;
              } else {
                lastDir->nextNode = tmpDir;
              } /* end if/else */
              lastDir = tmpDir;
#if DOTATS
              numDirs++;
#endif
#if UPLOADTRIG > 0
              /* If we found lots of dirs, link them up.  This should be a function, but... */
              uploadTrig++;
              if((lastDir != NULL) && (uploadTrig > UPLOADTRIG)) {
                uploadTrig = 0;
                pthread_mutex_lock(&dirList_mutex);
                /* If we already walked off the end... */
                if(gblWorkingDir == NULL) {
                  gblWorkingDir = firstDir;
                  gblLastDir = lastDir;
                } else {
                  gblLastDir->nextNode = firstDir;
                  gblLastDir = lastDir;
                } /* end if/else */
                firstDir = NULL;
                lastDir = NULL;
                pthread_cond_broadcast(&dirList_cond);
                pthread_mutex_unlock(&dirList_mutex);
              } /* end if */
#endif

#if DEBUG >= 10
              pthread_mutex_lock(&console_mutex);
              printf("=Local List=========================================\n");
              tmpDir = firstDir;
              while(tmpDir != NULL) {
                printf("%s\n", tmpDir->d_name);
                tmpDir = tmpDir->nextNode;
              } /* end while */
              printf("====================================================\n");
              pthread_mutex_unlock(&console_mutex);
#endif
            } else {
#if DOSTATS
              if(S_ISREG(s.st_mode)) {
                totalRegFileSize += s.st_size;
                totalRegFileBlks += s.st_blocks;
                numFiles++;
              } else if(S_ISLNK(s.st_mode)) {
                numLinks++;
              } else {
                numSpecial++;
              } /* end if/else */
#endif
            } /* end if/else */
#if DOSTATS
            /* These are added up for ALL files. */
            totalBlks += s.st_blocks;
            totalSize += s.st_size;
#endif
#if FILEACTION
            fileAction(s, name);
#endif
          } /* end if/else */
          free(name);
        } /* end if */
      } /* end if/else */
    } /* end while */
    closedir(dp);
    free(dep);
  } /* end if/else */

#if DOSTATS
  /* Update global stats variables now. */
  /* Only do it once to minimize mutex locking. */
  pthread_mutex_lock(&stats_mutex);
  gblNumDirs    += numDirs;
  gblNumFiles   += numFiles;
  gblNumLinks   += numLinks;
  gblNumSpecial += numSpecial;
  gblTotalBlks  += totalBlks;
  gblTotalSize  += totalSize;
  gblTotalRegFileSize += totalRegFileSize;
  gblTotalRegFileBlks += totalRegFileBlks;
  pthread_mutex_unlock(&stats_mutex);
#endif

  /* If we found dirs, link them up. */
  if(lastDir != NULL) {
    pthread_mutex_lock(&dirList_mutex);
    /* If we already walked off the end... */
    if(gblWorkingDir == NULL) {
      gblWorkingDir = firstDir;
      gblLastDir = lastDir;
    } else {
      gblLastDir->nextNode = firstDir;
      gblLastDir = lastDir;
    } /* end if/else */
    pthread_cond_broadcast(&dirList_cond);
    pthread_mutex_unlock(&dirList_mutex);
  } /* end if */

  /* Free up the structure and string for this dirent. */
  pthread_mutex_lock(&dirList_mutex);
  free(workingDir->d_name);
  free(workingDir);
  pthread_mutex_unlock(&dirList_mutex);
} /* end func readDir */
