/**
   @file      uidgidch.c
   @author    Mitch Richling <richmit@member.ams.org>
   @Copyright Copyright 1997,1998 by Mitch Richling.  All rights reserved.
   @brief     @EOL
   @Keywords  none
   @Std       C89

              Include file to create the uid changer.  Mixed the GID
              changer into the same header file 2002-3-28.  Now this
              header file can create two different classes of
              binaries.  One that changes just UIDs, and one that
              changes both UIDs and GIDs.
*/

#include <ctype.h>
#include <limits.h>

/* *************************************************************************** */
/* Functionality options */

/* Call a function one each file */
#define FILEACTION 1
#define INITACTION 1
#define EXITACTION 1

/* Exit on not fatal, bad arguments. */
#define BADARGEXIT 0

/* Trivial Errors cause an exit. */
#define TRIVERREXIT 0

/* Print the list as we process it */
#ifndef PRINTLIST
#define PRINLIST 0
#endif

/* Collect stastics */
#define DOSTATS 1

/* Set this to the name of the chown function to use.  Generally this
   will be one of 'chown' or 'lchown'.  This option exists to support
   platforms that do now have 'lchown' -- like some of the BSDs.*/
#ifndef USECHOWN
#define USECHOWN lchown
#endif

/* *************************************************************************** */
/* Tuning options */

/* Use memcpy instead of string functions. */
#define USEMEMCPY 0

/* Maximum number-1 of threads to run at one time. */
#define NUMTHREADS 50

/* Number of directories to read before updating global linked list. */
#define UPLOADTRIG 3

/* *************************************************************************** */
/* This is used for error messages. */
#if defined (dirgo_uidgidch)
#define PROGNAME "uidgidch"
#elif defined (dirgo_uidch)
#define PROGNAME "uidch"
#elif defined (dirgo_chown)
#define PROGNAME "chown"
#elif defined (dirgo_ltsstage)
#define PROGNAME "ltsstage"
#else
#define PROGNAME "ERROR"
#endif
/* *************************************************************************** */
/* Global variables defined by this module. These are global to the main.c as well. */

#if defined (dirgo_chown) || defined (dirgo_ltsstage)
long int uidForAll, gidForAll;
#endif
#if defined (dirgo_uidch) || defined (dirgo_uidgidch)
#define MAX_UID 0xfffff
#define MIN_UID 1
long int uidar[MAX_UID+1];
#endif
#if defined (dirgo_uidgidch)
#define MAX_GID 0xfffff
#define MIN_GID 1
long int gidar[MAX_GID+1];
#endif

#if defined (dirgo_ltsstage)
#define MY_SSIZE_MAX 16384
char *pathFrom;
char *pathTo;
int lenFrom;
int lenTo;
#if MY_SSIZE_MAX > SSIZE_MAX
  1/0;
#endif
#endif

time_t startTime;

/* Global variables defined by this module that are used for stastics. */
long int actGblNumUidChanges, actGblNumGoodUidChanges;
#if defined (dirgo_uidgidch) || defined (dirgo_chown) || defined (dirgo_ltsstage)
long int actGblNumGidChanges, actGblNumGoodGidChanges;
#endif
#if defined (dirgo_ltsstage)
long int actGblNumGoodLinkChanges, actGblNumLinkChanges;
#endif


/* Controls all the stats variables used by this code. */
pthread_mutex_t actionStats_mutex = PTHREAD_MUTEX_INITIALIZER;

#define LOGACTION 0

/* Set DOITFORREAL if not already set for us. */
#ifndef DOITFORREAL
#define DOITFORREAL 1
#endif

/* Set CUTOFFTIME if not already set for us. */
#ifndef CUTOFFTIME
#if defined(dirgo_chown) || defined(dirgo_ltsstage)
#define CUTOFFTIME 0
#else
#define CUTOFFTIME 1
#endif
#endif

time_t cutOffTime = 0;

/* *************************************************************************** */
/* The idea here is to chown a file if it is on the list. */
int fileAction(struct stat s, char *name)
{
  int r = 1;
#if defined (dirgo_ltsstage)
  char slinkPath[MY_SSIZE_MAX];
  char nlinkPath[MY_SSIZE_MAX];
  int linkWacked = 0;
#endif
  if( (cutOffTime == 0) || (s.st_ctime < cutOffTime) ) {

#if defined (dirgo_ltsstage)
    if(S_ISLNK(s.st_mode)) {
      r = readlink(name, slinkPath, MY_SSIZE_MAX);
      if (r < 0) {
        printf("ERROR(23): %s: readlink: %s\n", PROGNAME, "Undetermined");
      } else {
        if (r >= MY_SSIZE_MAX-2) {
          printf("ERROR(23): %s: link target too long: %s\n", PROGNAME, slinkPath);
        } else {
          slinkPath[r] = '\0';
          if(strncmp(pathFrom, slinkPath, lenFrom-1) == 0) {
            if ((lenTo+strlen(name)-lenFrom+16) > MY_SSIZE_MAX) {
              printf("ERROR(23): %s: New Link Too Long: %s\n", PROGNAME, name);
            } else {
              strncpy(nlinkPath, pathTo, lenTo+1);
              /* We know this is safe as we checked sizes above. */
              strcpy(&(nlinkPath[lenTo-1]), &(slinkPath[lenFrom-1]));
              pthread_mutex_lock(&actionStats_mutex); /* Count actions */
              actGblNumLinkChanges++;
              linkWacked = 1;
              pthread_mutex_unlock(&actionStats_mutex);
#if DOITFORREAL
              r = unlink(name);
#else
              r = 1;
#endif
#if(DEBUG > 4)
              pthread_mutex_lock(&console_mutex);
              printf("UNLINK(%d): %s\n", r, name);
              pthread_mutex_unlock(&console_mutex);
#endif
              if (r < 0) {
                printf("ERROR(23): %s: unlink: %s\n", PROGNAME, "Undetermined");
              } else {
#if DOITFORREAL
                r = symlink(nlinkPath, name);
#else
                r = 1;
#endif
#if(DEBUG > 4)
                pthread_mutex_lock(&console_mutex);
                printf("SYMLINK(%d): %s -> %s (WAS: %s)\n", r, name, nlinkPath, slinkPath);
                pthread_mutex_unlock(&console_mutex);
#endif
                if (r < 0) {
                  printf("ERROR(23): %s: symlink: %s\n", PROGNAME, "Undetermined");
                } else {
                  pthread_mutex_lock(&actionStats_mutex); /* Count actions that worked */
                  actGblNumGoodLinkChanges++;
                  pthread_mutex_unlock(&actionStats_mutex);
                }
              }
            }
          } else {
#if(DEBUG > 4)
            pthread_mutex_lock(&console_mutex);
            printf("LINKOK(%d): %s ::: %s -> %s\n", lenFrom, pathFrom, name, slinkPath);
            pthread_mutex_unlock(&console_mutex);
#endif
          }
        }
      }
    }
#endif

    if(
#if defined (dirgo_uidgidch)
      (s.st_gid >= MIN_GID) && (s.st_gid <= MAX_GID) && (s.st_uid >= MIN_UID) && (s.st_uid <= MAX_UID)
#elif  defined (dirgo_uidch)
      (s.st_uid >= MIN_UID) && (s.st_uid <= MAX_UID)
#elif  defined (dirgo_chown) || defined (dirgo_ltsstage)
      (1)
#endif
      ) {
#if dirgo_uidgidch
      long int newUid = uidar[s.st_uid];
      long int newGid = gidar[s.st_gid];
#elif defined (dirgo_uidch)
      long int newUid = uidar[s.st_uid];
      long int newGid = -1;
#elif defined (dirgo_chown)
      long int newUid = ( (s.st_uid != uidForAll) ? uidForAll : -1 );
      long int newGid = ( (s.st_gid != gidForAll) ? gidForAll : -1 );
#elif defined (dirgo_ltsstage)
      long int newUid = ( (linkWacked || (s.st_uid != uidForAll) ) ? uidForAll : -1 );
      long int newGid = ( (linkWacked || (s.st_gid != gidForAll) ) ? gidForAll : -1 );
#endif
      if( (newUid >= 0) || (newGid >= 0) ) {
        if(S_ISREG(s.st_mode) || S_ISDIR(s.st_mode) || S_ISLNK(s.st_mode)) {

#if DOITFORREAL
          r = USECHOWN(name, newUid, newGid);          
#else
          r = 1;
#endif
          if(r < 0) {
            pthread_mutex_lock(&console_mutex);
            printf("Could NOT chown file: %-s\n", name);
            pthread_mutex_unlock(&console_mutex);
          }
          if(newUid >= 0) {
#if LOGACTION
            pthread_mutex_lock(&console_mutex);
            printf("%2d UID %10ld=>%10ld  %-s\n", r, s.st_uid, newUid, name);
            pthread_mutex_unlock(&console_mutex);
#endif
            pthread_mutex_lock(&actionStats_mutex);
            actGblNumUidChanges++;
            if( !(r < 0)) 
              actGblNumGoodUidChanges++;
            pthread_mutex_unlock(&actionStats_mutex);
          }
#ifndef dirgo_uidch
          if(newGid >= 0) {
#if LOGACTION
            pthread_mutex_lock(&console_mutex);
            printf("%2d GID %10ld=>%10ld  %-s\n", r, s.st_gid, newGid, name);
            pthread_mutex_unlock(&console_mutex);
#endif
            pthread_mutex_lock(&actionStats_mutex);
            actGblNumGidChanges++;
            if( !(r < 0)) 
              actGblNumGoodGidChanges++;
            pthread_mutex_unlock(&actionStats_mutex);
          }
#endif
        } else {
          /* It might be a special file, but not a link. */
#if(DEBUG > 4)
          pthread_mutex_lock(&console_mutex);
          printf("NOT A REGULAR FILE OR DIR: %-s\n", name);
          pthread_mutex_unlock(&console_mutex);
#endif
        }
      } else {
#if(DEBUG > 4)
        pthread_mutex_lock(&console_mutex);
        printf("%2d %10ld=>%10s  %-s\n", r, (long int)s.st_uid, "N/A", name);
        pthread_mutex_unlock(&console_mutex);
#endif
      }
    } else { /* UID or GID was out of range -- only for dirgo_uidch & dirgo_uidgidch */
#if defined (dirgo_uidgidch) || defined (dirgo_uidch)
      if(s.st_uid > MAX_UID) {
        pthread_mutex_lock(&console_mutex);
        printf("UID %10ld out of range for %-s\n", (long int)s.st_uid, name);
        pthread_mutex_unlock(&console_mutex);
      }
#endif
#if defined (dirgo_uidgidch)
      if(s.st_gid > MAX_GID) {
        pthread_mutex_lock(&console_mutex);
        printf("GID %10ld out of range for %-s\n", (long int)s.st_gid, name);
        pthread_mutex_unlock(&console_mutex);
      }
#endif

    }
  }
  return (0);
}

/* *************************************************************************** */
/* This function loads a list of uids to change from the file named as
   the first argument to the application.  It also increments
   idxFirstArg so that the traversal routines will begin the
   traversal with the next argument.  The format of the file is
   "olduid=uid", one such pair per line. */
int initAction(int argc, char *argv[], int *idxFirstArg)
{
  int i;
#if CUTOFFTIME
  int cutOffArgNo;
#endif
#if defined (dirgo_uidgidch) || defined (dirgo_uidch) || CUTOFFTIME
char *tmpCharPtr;
#endif
#if defined (dirgo_uidgidch) || defined (dirgo_uidch)
char readBuff[1024], *tokenPoint;
  FILE *FD;
  long int fromID, toID;
#endif
#if defined (dirgo_chown) || defined (dirgo_ltsstage)
  char *p;
#endif
  int numArgsReq;

#if (DEBUG > 4)
  printf("DOITFORREAL=%d\n", DOITFORREAL);
#endif

  actGblNumUidChanges       = 0;
  actGblNumGoodUidChanges   = 0;
#if defined (dirgo_uidgidch) || defined (dirgo_chown) || defined (dirgo_ltsstage)
  actGblNumGidChanges       = 0;
  actGblNumGoodGidChanges   = 0;
#endif
#if defined (dirgo_ltsstage)
  actGblNumLinkChanges      = 0;
  actGblNumGoodLinkChanges  = 0;
#endif
#if defined (dirgo_uidgidch) || defined (dirgo_chown)
  numArgsReq = 4;
#elif defined (dirgo_ltsstage)
  numArgsReq = 6;
#elif defined (dirgo_uidch)
  numArgsReq = 3;
#endif
#if CUTOFFTIME
  numArgsReq++;
#endif
  *idxFirstArg = numArgsReq - 1;

  /* Print out the arguments... */
  for(i = 0; i < argc; i++)
    printf("%s ", argv[i]);
  printf("\n");

  if(argc < numArgsReq) {
    if ((argc>1) && (argv[1][0] == '-') && (argv[1][1] == 'h')) {
#if CUTOFFTIME

#if defined (dirgo_uidgidch)
      printf("INFO(0): %s: USE: %s UID_FILE GID_FILE CUTOFF_TIME DIR [DIR...]\n", PROGNAME, PROGNAME);
#elif defined (dirgo_chown)
      printf("INFO(0): %s: USE: %s UID_NUMBER GID_NUMBER CUTOFF_TIME DIR [DIR...]\n", PROGNAME, PROGNAME);
#elif defined (dirgo_ltsstage)
      printf("INFO(0): %s: USE: %s UID_NUMBER GID_NUMBER FROM_DIR TO_DIR CUTOFF_TIME DIR [DIR...]\n", PROGNAME, PROGNAME);
#elif defined (dirgo_uidch)
      printf("INFO(0): %s: USE: %s UID_FILE CUTOFF_TIME DIR [DIR...]\n", PROGNAME, PROGNAME);
#endif

#else

#if defined (dirgo_uidgidch)
      printf("INFO(0): %s: USE: %s UID_FILE GID_FILE DIR [DIR...]\n", PROGNAME, PROGNAME);
#elif defined (dirgo_chown)
      printf("INFO(0): %s: USE: %s UID_NUMBER GID_NUMBER DIR [DIR...]\n", PROGNAME, PROGNAME);
#elif defined (dirgo_ltsstage)
      printf("INFO(0): %s: USE: %s UID_NUMBER GID_NUMBER FROM_DIR TO_DIR DIR [DIR...]\n", PROGNAME, PROGNAME);
#elif defined (dirgo_uidch)
      printf("INFO(0): %s: USE: %s UID_FILE DIR [DIR...]\n", PROGNAME, PROGNAME);
#endif

#endif
    } else {
      printf("ERROR(12): %s: Not enough arguments (use -h for help)!\n", PROGNAME);
    }
    exit(1);
  }

#if defined(dirgo_ltsstage)
  if(argc > numArgsReq) {
    printf("ERROR(34): %s: too many arguments (use -h for help)!\n", PROGNAME);
    exit(1);
  }
#endif

#if CUTOFFTIME
  cutOffArgNo = numArgsReq - 2;
#if(DEBUG > 2)
  printf("cutOffTime(s): %s\n", argv[cutOffArgNo]);                    /* &&&& Date Ext &&&& */
#endif
  for(tmpCharPtr=argv[cutOffArgNo]; *tmpCharPtr!='\0'; tmpCharPtr++) 
    if( !(isdigit(*tmpCharPtr))) {
      printf("ERROR(42): %s: invalid CUTOFF_TIME: '%s'\n", PROGNAME, argv[cutOffArgNo]);
      exit(1);
    }
  if(strlen(argv[cutOffArgNo]) < 1) {
    printf("ERROR(41): %s: invalid CUTOFF_TIME: '%s'\n", PROGNAME, argv[cutOffArgNo]);
    exit(1);
  }
  cutOffTime = (time_t)(atol(argv[cutOffArgNo]));                              /* &&&& Date Ext &&&& */
#if(DEBUG > 2)
  printf("cutOffTime(i): %ld\n", (long)cutOffTime);                            /* &&&& Date Ext &&&& */
#endif
#endif

#if defined (dirgo_ltsstage)
  pathFrom = argv[3];
  pathTo   = argv[4];
  lenFrom  = strlen(pathFrom);
  lenTo    = strlen(pathTo);
  if(lenFrom <= 1) {
    printf("ERROR(39): %s: From path too short!\n", PROGNAME);
    exit(1);
  }
  if(lenTo <= 1) {
    printf("ERROR(40): %s: To path too short!\n", PROGNAME);
    exit(1);
  }
  if(pathFrom[0] != '/') {
    printf("ERROR(35): %s: From path must be fully qualified: %s!\n", PROGNAME, pathFrom);
    exit(1);
  }
  if(pathTo[0] != '/') {
    printf("ERROR(36): %s: To path must be fully qualified: %s!\n", PROGNAME, pathTo);
    exit(1);
  }
  if(pathFrom[lenFrom-1] != '/') {
    printf("ERROR(37): %s: From path must end with a /: %s!\n", PROGNAME, pathFrom);
    exit(1);
  }
  if(pathTo[lenTo-1] != '/') {
    printf("ERROR(38): %s: To path must end with a /: %s!\n", PROGNAME, pathTo);
    exit(1);
  }
#endif

  startTime = time(NULL);
  printf("Start Time: %ld\n", (long)startTime);

#if defined (dirgo_chown) || defined (dirgo_ltsstage)
  for(i=0, p=argv[1]; *p!=0; i++, p++)
    if ( !((i <= 7) && (((*p == 45) && (i == 0)) || ((*p >= 48) && (*p <= 57))))) {
      printf("ERROR(19): %s: UID was malformed or too large!: %s\n", PROGNAME, argv[1]);
      exit(1);
    }
  for(i=0, p=argv[2]; *p!=0; i++, p++)
    if ( !((i <= 7) && (((*p == 45) && (i == 0)) || ((*p >= 48) && (*p <= 57))))) {
      printf("ERROR(19): %s: GID was malformed or too large!: %s\n", PROGNAME, argv[2]);
      exit(1);
    }
  /* Now that we know the strings contain valid integers, we parse them. */
  uidForAll = atol(argv[1]);
  gidForAll = atol(argv[2]);
#endif

#if defined (dirgo_uidgidch) || defined (dirgo_uidch)
  for(i = MIN_UID; i <= MAX_UID; i++)
    uidar[i] = -1;
  /* Load the uid numbers from disk. */
  if((FD = fopen(argv[1], "r")) == NULL) {                  
    printf("ERROR(11): %s: open : %s\n", PROGNAME, argv[1]);
    exit(1);
  }
  while(NULL != fgets(readBuff, 1000, FD)) {
    if( (strlen(readBuff)>0) && (readBuff[strlen(readBuff)-1] == '\n')) {
      readBuff[strlen(readBuff)-1] = '\0';
    }
    tokenPoint = strchr(readBuff, '=');
    if(tokenPoint != NULL) {
      tmpCharPtr=readBuff;
      while(*tmpCharPtr != '\0') {
        if( !(isdigit(*tmpCharPtr)) && !(tmpCharPtr==tokenPoint) ) {
          printf("ERROR(12): %s: invalid UID map syntax for: '%s'\n", PROGNAME, readBuff);
          exit(1);
        }
        tmpCharPtr++;
      }
      fromID = atol(readBuff);
      toID = atol(tokenPoint + 1);
      if( (fromID >= MIN_UID) && (fromID <= MAX_UID) ) {
        if( (toID >= MIN_UID) && (toID <= MAX_UID) ) {
          uidar[fromID] = toID;
#if(DEBUG > 4)
          printf("UID MAP: %10ld=>%-10ld\n", fromID, toID);
#endif
        } else {
          printf("RHS UID was out of range(%ld)\n", toID);
          exit(29);
        }
      } else {
        printf("LHS UID was out of range(%ld)\n", fromID);
        exit(28);
      }
    } else {
      printf("Error processing(missing equals): '%s'\n", readBuff);
      exit(13);
    }
  }
  fclose(FD);
#endif

#if defined (dirgo_uidgidch)
  for(i = MIN_GID; i <= MAX_GID; i++)
    gidar[i] = -1;
  /* Load the gid numbers from disk. */
  if((FD = fopen(argv[2], "r")) == NULL) {                  
    printf("ERROR(11): %s: open : %s\n", PROGNAME, argv[2]);
    exit(1);
  }
  while(NULL != fgets(readBuff, 1000, FD)) {
    if( (strlen(readBuff)>0) && (readBuff[strlen(readBuff)-1] == '\n')) {
      readBuff[strlen(readBuff)-1] = '\0';
    }
    tokenPoint = strchr(readBuff, '=');
    if(tokenPoint != NULL) {
      tmpCharPtr=readBuff;
      while(*tmpCharPtr != '\0') {
        if( !(isdigit(*tmpCharPtr)) && !(tmpCharPtr==tokenPoint) ) {
          printf("ERROR(12): %s: invalid GID map syntax for: '%s'\n", PROGNAME, readBuff);
          exit(1);
        }
        tmpCharPtr++;
      }
      fromID = atol(readBuff);
      toID = atol(tokenPoint + 1);
      if( (fromID >= MIN_GID) && (fromID <= MAX_GID) ) {
        if( (toID >= MIN_GID) && (toID <= MAX_GID) ) {
          gidar[fromID] = toID;
#if(DEBUG > 4)
          printf("GID MAP: %10ld=>%-10ld\n", fromID, toID);
#endif
        } else {
          printf("RHS GID was out of range(%ld)\n", toID);
          exit(32);
        }
      } else {
        printf("LHS GID was out of range(%ld)\n", fromID);
        exit(31);
      }
    } else {
      printf("Error processing(missing equals): '%s'\n", readBuff);
      exit(13);
    }
  }
  fclose(FD);
#endif
  fflush(stdout);
  fflush(stderr);
  return (0);
}

/* *************************************************************************** */
/* This is meant to be overridden with useful stuff. */
int exitAction(void)
{
  printf("\n");
  printf("Number of regular files... %ld\n", gblNumFiles);
  printf("Number of links........... %ld\n", gblNumLinks);
  printf("Number of directories..... %ld\n", gblNumDirs);
  printf("Number of special files... %ld\n", gblNumSpecial);
  printf("Number uid changes........ %ld/%ld\n", actGblNumGoodUidChanges, actGblNumUidChanges);
#if defined (dirgo_uidgidch) || defined (dirgo_chown) || defined (dirgo_ltsstage)
  printf("Number gid changes........ %ld/%ld\n", actGblNumGoodGidChanges, actGblNumGidChanges);
#endif
#if defined (dirgo_ltsstage)
  printf("Number link changes....... %ld/%ld\n", actGblNumGoodLinkChanges, actGblNumLinkChanges);
#endif
  printf("Seconds................... %ld\n", (long)(time(NULL)-startTime));
  return 0;
}
